# -*- coding: utf-8 -*-


import time, math

from collections import defaultdict

import _feature_  as feature
import _read_xml_ as read_xml
import _conf_     as conf
import _perceptron_ as p


# -----------------------------------------------------------------------------
# averaging weights. (binary)
# -----------------------------------------------------------------------------
def averageWeights(weights, w_0, w_a, data_num):
    for key in weights.keys():
        weights[key] = w_0[key] - w_a[key]/data_num


# -----------------------------------------------------------------------------
# get output of binary classification with averaged-perceptron.
# [INPUT]
#     inputs   : input vector (float dict)
#     weights  : weight vector (float dict)
#     gold     : gold label (boolean)
#     w_0      : sum_0_n{delta}
#     w_a      : sum_0_i{sum_i_n{delta}}
#     data_num : for average
# [OUTPUT]
#     (bool) : whether are weights updated
# -----------------------------------------------------------------------------
def getNewWeights(inputs, weights, w_0, w_a, gold, data_num):
    averageWeights(weights, w_0, w_a, data_num)
    output = getOutput(inputs, weights)

    if (output > 0) != gold:
        t = +1 if gold else -1

        for key, w in weights.iteritems():
            if inputs.has_key(key):
                delta = inputs[key] * t * p.eta
                w_0[key] += delta
                w_a[key] += delta * data_num

        do_update = True
    
    else:
        do_update = False
    
    return do_update


# -----------------------------------------------------------------------------
# get output of binary classification with averaged-perceptron.
# [INPUT]
#     inputs  : input vector (float dict)
#     weights : weight vector (float dict)
# -----------------------------------------------------------------------------
def getOutput(inputs, weights):
    output = 0
    
    for key, x in inputs.iteritems():
        output += weights[key] * x

    return output


# -----------------------------------------------------------------------------
# output weights as model-data.
# -----------------------------------------------------------------------------
def writeWeights(filename, weights):
    with open(filename, 'w') as fo:
        for key, w in weights.iteritems():
            fo.write('%s:%lf\n' % (key, w))


# -----------------------------------------------------------------------------
# load model-data.
# -----------------------------------------------------------------------------
def loadWeights(filename):
    weights = defaultdict(float)
    
    with open(filename) as fi:
        for line in fi:
            spl = line.rsplit(':', 1)
            if spl:
                weights[spl[0]] = float(spl[1])

    return weights


# -----------------------------------------------------------------------------
# convert data to input-vector and gold-label.
# -----------------------------------------------------------------------------
def convertData(data_list, filename_feat):
    # output the feature file
    feature.writeFeatures(filename_feat, data_list)

    inputs_list = []
    gold_list   = []

    # parse .f file
    with open(filename_feat) as fi:
        for i, line in enumerate(fi):
            spl = line.split()
            gold = (spl[0] != '-1')
            x = dict()

            for s in spl[1:]:
                t = s.rsplit(':', 1)
                if len(t) == 2:
                    x[t[0]] = float(t[1])
                else:
                    x[s] = 1.0
            
            inputs_list.append(x)
            gold_list.append(gold)

    return inputs_list, gold_list


# -----------------------------------------------------------------------------
# train on data_list.
# -----------------------------------------------------------------------------
def train(data_list):
    weights = defaultdict(float)
    w_0     = defaultdict(float)
    w_a     = defaultdict(float)

    print 'FN_DEV_F =', p.fn_dev_f
    print 'FN_TEST_F =', p.fn_test_f
    print 'FN_MODEL_F =', p.fn_model
    
    inputs_list, gold_list = convertData(data_list, p.fn_dev_f)

    print 'Number of instance:', len(inputs_list)
    print ''

    print 'max_iterations:', p.max_iter
    print 'period:',         p.period
    print 'epsilon:',        p.epsilon
    print 'eta:',            p.eta
    print ''

    # training
    training_start_time = time.time()
    data_num  = 1
    end_count = 0
    loss_log  = []
    
    for i in xrange(p.max_iter):

        print '***** Iteration #%d *****' % (i+1)
        
        start_time = time.time()
        loss = 0
        
        for j, inputs in enumerate(inputs_list):
            if getNewWeights(inputs, weights, w_0, w_a, gold_list[j], data_num):
                loss += 1
            data_num += 1

        end_time = time.time()

        if loss == 0:
            end_count += 1
        else:
            end_count = 0
        
        print 'Loss:', loss
        print 'Feature L2-norm:', sum([w*w for w in weights.values()])
        print 'Seconds required for this iteration: %lf\n' % (end_time-start_time)

        # judge convergence
        if i >= p.period:
            variance = p.getVariance(loss_log)
            
            if variance/max(1,loss) < p.epsilon:
                print 'Terminated with the stopping criterion'
                break

        loss_log.append(loss)
        if len(loss_log) > p.period:
            del loss_log[0]

    training_end_time = time.time()

    print 'Seconds required: %lf' % (training_end_time - training_start_time)

    averageWeights(weights, w_0, w_a, data_num)
    writeWeights(p.fn_model, weights)


# -----------------------------------------------------------------------------
# test on data_list.
# [INPUT]
#     data_list : target to test (Pair list)
# -----------------------------------------------------------------------------
def test(data_list):
    print 'start test:'
    print '-' * 40
    
    weights = loadWeights(p.fn_model)
    inputs_list, gold_list = convertData(data_list, p.fn_test_f)

    print 'Number of instance:', len(inputs_list)
    print ''

    results = []
        
    for i, inputs in enumerate(inputs_list):
        out = getOutput(inputs, weights)
        prob_id = data_list[i].attr['id']
        label = 'Y' if out>0 else 'N'

        if conf.do_apply_sigmoid:
            score = 1.0 / (1.0 + math.exp(-1.0 * out))
            if label == 'N':
                score = 1.0 - score

        results.append('%s %s %lf' % (prob_id, label, out))

    print '-' * 40
    print 'complete test.'

    results_str = '\n'.join(results)

    with open(p.fn_test_log, 'w') as fo:
        fo.write(results_str)

    return results_str


# -----------------------------------------------------------------------------
# do cross-validations.
# -----------------------------------------------------------------------------
def crossvalidate():
    # make feature files
    data_list = conf.getDevData()
    size      = len(data_list) / conf.sep_num

    print 'start cross-validation :'
    print '-' * 40

    result_list = []

    for i in xrange(conf.sep_num):
        print 'cross-validation (%d/%d):' % (i+1, conf.sep_num)
        print '-' * 40

        idx = size * (i+1) if i < conf.sep_num-1 else len(data_list)
        
        data_list_dev  = data_list[:size*i] + data_list[idx:]
        data_list_test = data_list[size*i:idx]

        dn = conf.log_dir_name
        p.fn_dev_f    = '%s/cross_%d_dev.f'    % (dn, i)
        p.fn_test_f   = '%s/cross_%d_test.f'   % (dn, i)
        p.fn_model    = '%s/cross_%d.model'    % (dn, i)
        p.fn_dev_log  = '%s/cross_%d_dev.log'  % (dn, i)
        p.fn_test_log = '%s/cross_%d_test.log' % (dn, i)
        
        train(data_list_dev)
        result = test(data_list_test)
        result_list.append(result)

        print '-' * 40

    print '-' * 40
    print 'complete cross-validation.'

    fn_log = '%s/cross_test_all.log' % conf.log_dir_name
    with open(fn_log, 'w') as fo:
        fo.write('\n'.join(result_list))

